"""
For a network, every single node is taken, its mask and state vector is clipped.
They are subject to a boolean operation from numpy libraries.
Then from the fact that whether the node outputs a number greater than the 
shortened state vector, the node outputs 1 or 0.
These values are gathered up and outputted as a new state vector.
The method can be:
    num.logical_or
    num.logical_and
    num.logical_xor
    num.logical_xnor
    
    or any other function that outputs an integer with an input of two ndarrays.        
"""

from boolfuncs import *





from kreveik import *
import numpy as num
import copy
import itertools 
import logging

def random_killer(ensemble,times):
    logging.info("Killing "+str(times)+" individuals")
    for i in range(times):
        randomnum = num.random.randint(0,len(ensemble))
        logging.info("("+str(i)+"/"+str(times)+") Killing "+str(ensemble[randomnum]))
        ensemble.remove(randomnum)
    
def qualified_killer(ensemble,**kwargs):
    pass

def underachiever_killer(ensemble,**kwargs):
    pass

    


"""
Probes are objects that are meant to calculate and accumulate data from the
subroutines of the  simulation.
"""

import numpy as num 

class SubRoutine(object):
    def __init__ (self):
        pass
        
    
genetic_algorithm = SubRoutine()
genetic_iteration = SubRoutine()
populate_wildtype = SubRoutine()
populate_equilibria_in_family = SubRoutine()
populate_equilibria = SubRoutine()
search_equilibrium = SubRoutine()
advance = SubRoutine()


class probe (object):
    """
    Definition of the probe 
    A probe is has an assignee, and it is defined by listing a probe
    in the list "probes" within that object. (it can be a family or
    a network, obviously)
    The function is defined due to the Network 
    """
    def __init__ (self):
        self.function = None
        self.subroutine = None
        self.data = []

class family_mean_score_probe (probe):
    def __init__ (self):       
        probe.__init__(self)
        self.function = self.mean_score
        self.subroutine = genetic_iteration
        
    def mean_score(self,probeable):
        return num.mean(probeable.scores)
    
class mean_orbit_length_probe (probe):
    def __init__ (self):       
        probe.__init__(self)
        self.function = self.eq_score
        self.subroutine = populate_equilibria
        
    def eq_score(self,probeable):
        return [num.mean(probed.equilibria) for probed in probeable.network_list]
    
class motif_network_probe (probe):
    def __init__ (self,degree):       
        probe.__init__(self)
        self.function = self.motif_extractor
        self.subroutine = genetic_iteration
        self.degree = degree
        
    def motif_extractor(self,probeable):
        motif_out = probeable.motif_freqs(probeable,self.degree)
        return motif_out
    
class motif_family_probe (probe):
    def __init__ (self,degree,(first,last)):       
        probe.__init__(self)
        self.function = self.motif_extractor
        self.subroutine = genetic_iteration
        self.degree = degree
        self.first = first
        self.last = last
        
    def motif_extractor(self,probeable):
        return [network.motif_freqs(self.degree) for network 
                in probeable.network_list[self.first:self.last]]
        



"""
Definition of network object.
"""

import numpy as num
import matplotlib.pyplot as plt
import copy
import itertools 
import logging
from kreveik.classes import *
import kreveik.probes as probes
from kreveik import *


class TopologicalNetwork(ProbeableObj):
    """
    This object is a stripped down network, designated to be a core 
    object for all network-like objects, like sub-graphs and motifs.
    """
    def __init__ (self,adjacency_matrix):
        ProbeableObj.__init__(self)
        self.adjacency = num.array(adjacency_matrix,dtype=bool)
        self.code = str(len(self.adjacency))+"-"+str(reduce(lambda x,y : 2*x+y, 
                                                              self.adjacency.flatten()*1))
    
    def indegree(self):
        return self.adjacency.sum(axis=0)
    
    def outdegree(self):
        return self.adjacency.sum(axis=1)
    
    def graph(self):
        """Returns a pydot graph object.    
        """
        import pydot
        
        node_orig = 1
    
        graph = pydot.Dot(graph_type='digraph')

        for row in self.adjacency:
            skip = 0
            r = row
            node_dest = skip+1
    
            for e in r:
                if e:
                    graph.add_edge(
                        pydot.Edge( node_orig, node_dest) )
                node_dest += 1
            node_orig += 1
    
        return graph

        
        
    def is_connected(self):
        """
        Returns True if the graph is connected, False if not.
        uses the algorithm explained in 
        http://keithbriggs.info/documents/graph-eigenvalues.pdf
        """
        symmetric = self.adjacency+self.adjacency.T-num.diag(
                                 self.adjacency.diagonal())
        if (0 in symmetric.sum(axis=0) or 0 in symmetric.sum(axis=1)):
            return False
        degrees = num.diagflat(symmetric.sum(axis=0))
        laplacian = degrees-symmetric
        determinant = num.linalg.det(laplacian +num.ones((len(laplacian),len(laplacian) )))
        return not(determinant == 0)
    
    def copy(self):
        """
        Returns a copy of the Topological Network object. 
        """
        return copy.deepcopy(self)
    
    def save(self,filename):
        """
        Saves the Network as an object to a file specified.
        """
        import pickle
        try:
            filehandler = open(filename+".net", 'w')
            pickle.dump(self,filehandler) 
        except pickle.PickleError:
            logging.error("The object failed to be pickled.")

            

class Motif(TopologicalNetwork):
    """
    Motif is a 
    """
    def __init__(self, adjacency_matrix):
        TopologicalNetwork.__init__(self, adjacency_matrix)
        self.degree = len(adjacency_matrix)
    
    def __eq__(self,other):
        permutation_list = itertools.permutations(range(self.degree),self.degree)
        for permutation in permutation_list:
            
            if num.sum(self.indegree()) != num.sum(other.indegree()):
                return False

            newarray = num.zeros((len(self.adjacency),len(self.adjacency)),dtype=bool)
            #newarray[[node_init,node_end]] = newarray[[node_end,node_init]]
            #newarray[:,[node_init,node_end]] = newarray[:,[node_end,node_init]]
            for rowctr,row in enumerate(self.adjacency):
                for colctr,col in enumerate(row):
                    if col == True:
                        newarray[permutation[rowctr]][permutation[colctr]]= True
        
            if num.all(newarray == other.adjacency):
                return True
            
        return False    


class Network(TopologicalNetwork,Element):
    '''
    Network Class
    
    Input Arguments
        adjacency_matrix
        mask
        state_vec  
    '''
    def __init__ (self,adjacency_matrix,mask,function,state_vec=None):
        Element.__init__(self)
        TopologicalNetwork.__init__(self,adjacency_matrix)
        self.n_nodes= num.size(adjacency_matrix,0)
        self.mask=mask
        if state_vec == None:
            state_vec= (num.random.random((1,self.n_nodes))< 0.5)
        self.state=num.array(state_vec)
        self.function = function
    
    def __str__(self):
        return str(id(self))
        
    def info(self):
        '''
        Prints out an identification of the Network.
        Prints:
            Id
            Mothers
            Children
            Orbits
            Score
            Adjacency matrix
            sTate
            masK
        '''
        print "This network is : "+str(id(self))+"."
        print "Nodes: "+str(self.n_nodes)
        print "Score: "+str(self.score)
        print "Its mothers are: "
        print "   "+str(self.mother)
        print "Its children are: "
        for child in self.children:
            print "   "+str(child)
        print "It has the following adjacency matrix: "
        print self.adjacency
        print "The following are the masks for each node: "
        for (num,node) in enumerate(self.mask):
            print str(num)+" th node : "+str(node)
        print "The following are the states with respect to time "
        for (t,state) in enumerate(self.state):
            print "t= "+str(t)+" : "+str(node)
        print "The scorer is : "
        print self.scorer
        
    def __getitem__(self, index):
        """
        nth item of a network object is the state that it is in, in the nth 
        iteration
        """
        if index > len(self.state):
            raise IndexError
        return self.state[index]
    
    def __contains__(self, state):
        """
        Returns a boolean according to whether a network includes the state  
        """
        item = num.array(state*True)
        return item in self.state
            
    def __call__ (self,state):
        """
        When a  network is called as a function, it sets the initial condition 
        as the given vector, finds the equilibrium of that state.
        """
        self.set_state(state)
        self.search_equilibrium(2**self.n_nodes,state,orbit_extraction=False,def_advance=1)
        
        
    def advance(self,times,start_from=None,*args):
        '''
        Advances the state in the phase space a given number of times.
        If a starter state is given, the initial condition is taken as the given state.
        If not, the last state is used instead.
        Input Arguments
            times -> the number of iterations to be taken.
            starter_state -> the initial state to be used
        '''
        
        if start_from != None:
            self.set_set(start_from)
        
        for counter in xrange(times):
            newstate = self.function(self,self.state[-1])
            self.state = num.append(self.state,[newstate],axis=0)
            
        self.populate_probes(probes.advance)
        
    def set_state(self,state):
        """
        Flushes the state of the system, and sets the new state as the given one 
        """
        
        if type(state) == int:
            state = [int(strings)==True for strings in list(num.binary_repr(
                                                (state),width=self.n_nodes))]
        state_bool = [i == True for i in state]
        state = [list(state_bool)]
        self.state = num.array(state)
        
    def plot_state(self,last=20):
        '''
        Plots the last 20 states as a black and white strips vertically.
        The vertical axis is time, whereas each strip is a single state.
        Input Arguments
            last -> the number of states that will be plotted 
        '''
        # Take the state vector, convert the list of arrays into a 2d array, then show it as an image
        # Black and white. 
        
        plt.imshow(self.state[-last:],cmap=plt.cm.binary,interpolation='nearest')     
        plt.show()


    def plot_equilibria(self):
        """Creates a plot of the equilibria for all possible initial conditions
        in the phase space. Every point in the phase space corresponds to the 
        length of the orbit that initial condition is attracted to.
        """
        rowsandcols = 2**(len(self.adjacency)/2)
        if self.n_nodes % 2 == 0:
            im_matrix = self.equilibria.reshape((rowsandcols,rowsandcols))
        
        if self.n_nodes % 2 == 1:
            im_matrix = self.equilibria.reshape((rowsandcols,rowsandcols*2))
    
        plt.imshow(im_matrix,cmap=plt.cm.gray,interpolation='nearest')
        
        plt.grid()
        plt.colorbar()
        plt.show()
             
           
    def search_equilibrium(self,chaos_limit,starter_state,orbit_extraction=False,def_advance=1):
        '''
        Searches for an equilibrium point, or a limit cycle. 
        Returns the state vector, or the state vector list, if the equilibrium is a limit cycle.
        If no equilibrium is found, returns False.
        Input Arguments:
            starter_state -> the initial state vector that the state will evolve on.
            chaos_limit -> the degree that an orbit will be considered as chaotic.
                The calculation will stop when this point is reached.
            orbit_extraction -> True when every individual orbit is recorded with its degree.
        '''
        if not(hasattr(self,"equilibria")):
            self.equilibria = num.zeros(2**self.n_nodes)
        if not(hasattr(self,"orbits")):
            if orbit_extraction:
                self.orbits = num.array([None]*2**self.n_nodes)
        
        self.set_state(starter_state)
        starter_state = self.state[-1]
        
        for ctr in xrange(chaos_limit):
            
            self.advance(def_advance)
            row = num.all(self.state[-1] == self.state, axis=1) 
            where = num.where(row==True)
            
            if len(where[0])> 1:
                frst_where = where[0][0]
                scnd_where = where[0][1]
                
                orbit_length = scnd_where-frst_where
                
                orbit = None
                location = reduce(lambda x,y : 2*x+y, starter_state)
                
                if orbit_extraction:
                    orbit = self.state[frst_where:scnd_where]
                    self.orbits[location] = orbit
                 
                self.equilibria[location] = orbit_length
                
                self.populate_probes(probes.search_equilibrium)
                return (orbit_length,orbit)
        
        
            
    def populate_equilibria(self,orbit_extraction=False):
        '''
        Creates all possible initial conditions by listing all possible 2^n boolean states.
        Then runs populate_equilibrium for each of them.
            populate_equilibrium returns orbits and-or degrees of the orbits.
        Gives scores to each of the networks, depending on the degree of the orbit each initial condition
        rests on.
        Input Arguments:
            normalize -> normalizes the scores to the value given.
        '''
        
        self.equilibria = num.zeros(2**self.n_nodes)
        if orbit_extraction:
            self.orbits = num.array([None]*2**self.n_nodes)
        
        binspace = range(0,num.power(2,self.n_nodes))
  
        for state in binspace:
            self.search_equilibrium(2**self.n_nodes,state,orbit_extraction)  
        self.populate_probes(probes.populate_equilibria)

    




"""
Kreveik is a package that is meant to provide access to a set of tools 
about 
"""

import classes
import numpy as num
import matplotlib.pyplot as plt
import network
import probes
import copy
import genetic
import family
import parallel
import logging
logging.basicConfig(level=logging.INFO)

__all__ = ['classes','family','genetic','network','probes']




import numpy as num
from kreveik.classes import Network
import logging

def random(n_nodes,function,probability=(0.5,0.5,0.5),connected=False):
    '''
    Generates and returns a random network with a random initial conditions.
    The adjacency matrix, initial state, boolean function are populated with 
    1's and 0's with probabilities of 0.5.
    
    Input Variables
    n_nodes -> Number of nodes in the system. A value of 2 to 8 will generate a 
    network with manageable size and complexity.
    probability -> The probability of having a connection in any two nodes. 
    A smaller value will decrease complexity and clustering coefficient.  
    scorer -> a function for the type network
    '''

    num.random.seed()
    adjacency_matrix=(num.random.random((n_nodes,n_nodes))<probability[0])
    state=(num.random.random((1,n_nodes))<probability[1])
    bool_fcn=(num.random.random((n_nodes,n_nodes))<probability[2])
    new_network = Network(adjacency_matrix, bool_fcn, function, state_vec=state)
    
    if connected==False:
        try:
            logging.info("Generating one network of node count "+str(n_nodes))
            return new_network
        
        except ValueError,e:
            z = e
            logging.error("Network is too big to model.")
            print z
            
    elif new_network.is_connected():
        try:
            logging.info("Generating one network of node count "+str(n_nodes))
            return new_network
        
        except ValueError,e:
            z = e
            logging.error("Network is too big to model.")
            print z
    else:
        while not new_network.is_connected():
            num.random.seed()
            adjacency_matrix=(num.random.random((n_nodes,n_nodes))<probability[0])
            state=(num.random.random((1,n_nodes))<probability[1])
            bool_fcn=(num.random.random((n_nodes,n_nodes))<probability[2])
            new_network = Network(adjacency_matrix, bool_fcn, function,state_vec=state)
        
        try:
            logging.info("Generating one network of node count "+str(n_nodes))
            return new_network
        except ValueError,e:
            z = e
            logging.error("Network is too big to model.")
            print z
         
    return new_network


"""
In this module, the selectors for the GA elements, networks are defined. A selector returns True
or False due to whether a particular element is eligible for mutation or not. 
"""

import math
import numpy as num
import kreveik.classes as classes
import kreveik.genetic as genetic
import logging

def hard_threshold(network,**kwargs):
    if not('threshold' in  kwargs):
        logging.error("The hard_threshold Selector needs a threshold parameter to work.")
        return None
    if network.score < kwargs['threshold']:
        logging.info("score = "+str(network.score)+" < "+str(kwargs['threshold'])+" = threshold.")
        return True
    elif network.score == kwargs['threshold']:
        logging.info("score = "+str(network.score)+" = "+str(kwargs['threshold'])+" = threshold.")
        if num.random.randint(0,2) == 1:
            return True
        else:
            return False
    else:
        return False
def hard_threshold_with_probability(network,**kwargs):
    if not('threshold' in kwargs):
        logging.error("The hard_threshold Selector needs a threshold parameter to work.")
        return None
    if not('prob' in kwargs):
        logging.error("The hard_threshold Selector needs a prob parameter to work.")
        return None
    
    if network.score < kwargs['threshold']:
        logging.debug("score = "+str(network.score)+" < "+str(kwargs['threshold'])+" = threshold.")
        if num.random.random() < kwargs['prob']:
            return True
        else:
            return False
    elif network.score == kwargs['threshold']:
        logging.debug("score = "+str(network.score)+" = "+str(kwargs['threshold'])+" = threshold.")
        if num.random.randint(0,2) == 1:
            return True
        else:
            return False
    else:
        return False
    
def logistic(network,**kwargs):
    if not ('angle' in kwargs) or not('midpoint' in kwargs):
        logging.error("The logistic Selector needs angle and midpoint parameters to work.")
        return None
    angle = kwargs['angle']
    midpoint = kwargs['midpoint']
    prob = 1-(1/(1+math.exp(-angle*(network.score-midpoint))))
    if prob < num.random.random():
        return True
    else:
        return False




'''
The genereg package maintains class definitions and methods for manipulation,
statistical evaluation of random boolean networks. The emphasis is upon genetic
algorithm and how networks evolve accordingly.
'''

import numpy as num
import matplotlib.pyplot as plt
import logging

class ProbeableObj (object):
    """Base class for objects that a Probe object can be attached. 
    """
    def __init__ (self):
        self.probes = []
    def attach (self,probe):
        if probe in self.probes:
            logging.warning('The probe is already attached.')
        else:
            self.probes.append(probe)
    def __call__(self,probed):
        self.attach(probed)
         
    def populate_probes (self,subroutine):
        for probe in self.probes:
            if probe.subroutine == subroutine:
                measured = probe.function(self)
                probe.data.append(measured)
                
class Ensemble(object):
    """Base class for all that are going to be used as an ensemble of elements that are subject to 
    Genetical Algorithm
    """
    def __init__(self):
        self.scorer = None
        self.selector = None
        self.mutator = None
        self.killer = None
        
class Element(object):
    """Base class for all that are going to be used as individuals within an element in a Genetical 
    Algorithm
    """
    def __init__(self):
        self.score = None
        self.mutated = 0
        self.parent = None
        self.children = []


from network import TopologicalNetwork,Motif,Network
from family import Family





"""
Definition of the family class
"""
import numpy as num
import matplotlib.pyplot as plt
import logging
from kreveik.classes import *
import kreveik.probes as probes
from kreveik import parallel


class Family(ProbeableObj,Ensemble):
    '''
    Family Class.
    A class that will have a single family of networks, a genetic algorithm 
    will be run within a family. The macro parameters of a group of network, 
    though may also be reached from a list of networks, will be housed in a 
    family in a more structured manner.
    '''
    
    def __init__ (self):
        ProbeableObj.__init__(self)
        Ensemble.__init__(self)
        self.network_list = []
        self.wildtype_list = [] 
        self.scores = num.array([])
        self.scores_history = []
        
            
    def __getitem__(self, index):
        """
        nth item of a Family object returns the nth network it has
        """
        if index > len(self.network_list):
            raise IndexError
        return self.network_list[index]
    
    def __contains__(self, network):
        """
        Returns a boolean according to whether a family includes a network  
        """
        return network in self.network_list
        
    def __len__ (self):
        return len(self.network_list)
    
    def remove(self,n):
        self.network_list = num.delete(self.network_list, n).tolist()
                
    def add(self, network):
        '''
        This method will add a network to the specified family.
        network -> The network that is to be appended to the family.
        '''
        
        if network in self.network_list:
            logging.warning("The network that you're trying to add to the family is already a member.\
            Please clone the individual and append the clone.")
            return False
        
        self.network_list.append(network)
        return True
            
    def plot_scores(self):
             
        plt.plot(self.scores)
        plt.show()
    
    def populate_equilibria(self,pp=False):
        '''
        If the family has individuals, it goes to each individual and finds the equilibria 
        for all possible initial conditions they may face. The orbits and scores are 
        assigned to each one of them.
        '''
        
        if len(self.network_list) == 0:
            logging.warning("Warning: There is nobody in this family.Please adopt individuals.")
            return False
            
        self.scores = num.zeros(len(self.network_list))
        if pp:
            
            jobs = [(network,parallel.JOBSERVER.submit(network.populate_equilibria,(), 
            (self.equilibria, self.search_equilibrium, self.scorer, self.populate_probes),
                                              ("numpy as num",))) for network in self]
            for (ctr,(network,job)) in enumerate(jobs):
                job()
                self.scores[ctr] = network.score
        else:    
            for counter,network in enumerate(self.network_list): 
                logging.info("("+str(counter+1)+"/"+str(len(self.network_list))
                             +") Populating equilibrium for: "+str(network))
                network.populate_equilibria()
                self.scores[counter] = network.score
        self.populate_probes(probes.populate_equilibria_in_family)
            





import logging


def score(element,scorer=None):
    """
    If a scorer is specified, the score of the element calculated with that
    scorer is returned. 
    If scorer is not set, then the element is scored with the scorer specified 
    in its definition.
    """
    if scorer == None:
        element.score = element.scorer(element)
    else:
        return scorer(element)

class Mutator (object):
    """This is the class definition for the mutators. A mutating function must be given as an
    input
    """
    def __init__ (self, function):
        self.mutation = function
    def __call__ (self,element):
        self.mutation(element)
    
        
class Selector (object):
    """This is the class definition for the selector of the genetic iteration
    The selecting function must be supplied as an input.
    """
    def __init__ (self,function,**kwargs):
        self.selection = function
        self.kwargs = kwargs
    def __call__ (self,element):
        self.selection(element,self.kwargs)  

class Killer (object):
    """This is the base class definition for the killer. A killing function must be supplied, which
    will act on ensemble objects.
    """ 
    def __init__(self,function):
        self.scoring=function
    def __call__(self,ensemble):
        self.scoring(ensemble)
        

class Scorer (object):
    """This is the base class definition for the scorers. A scoring function must be supplied.
    """ 
    def __init__(self,function):
        self.scoring=function
    def __call__(self,element):
        self.scoring(element)
        

def genetic_iteration(ensemble,**kwargs):
    '''
    Runs one iteration of the genetic algorithm.
    It finds wildtypes of the family, mutates them, populates the family with mutants
    and assasinates as much of it has mutated. 
    '''
    if ((ensemble.scorer == None) or (ensemble.selector == None)
        or (ensemble.mutator == None) or (ensemble.killer == None)):
        raise ValueError("An element needs its scorer, killer, selector and mutator \
        defined in order to be fed into the GA")
        return False
    
    logging.info("GA: for ensemble "+str(ensemble)+" started.")
    killcount = 0 
    
    newcomer_list = []
    for element in ensemble:
        try:
            element.score = ensemble.scorer(element)
            logging.info("GA: Scoring Network "+str(element))
        except: 
            logging.error("GA: The scoring of the element failed.")
        
        if ensemble.selector(element,**kwargs):
            
            newcomer = element.copy()
            logging.info("GA: Mutating Wildtype "+str(element))
            ensemble.mutator(newcomer)
            
            newcomer_list.append(newcomer)
            killcount += 1
            
    
    logging.info("GA: Adding Newcomers")
    for individual in newcomer_list:
        ensemble.add(individual)
        logging.info("New network added, "+str(individual)+".")
        individual.populate_equilibria()
        individual.score = ensemble.scorer(individual)
    
    logging.info("GA: Killing...")
    ensemble.killer(ensemble,killcount)





"""
This module contains different scorer functions that have a 
single network as input and outputs its score accordingly.
"""

import numpy as num
 
def sum_scorer_f(network):
    """
    This function takes a network object and returns its score 
    computed by summing the lengths of orbits for every single 
    possible initial condition. If initial conditions turn 
    out to have the same attractors, they are counted again.
    """
    network.populate_equilibria()
    return sum(network.equilibria)/(2.0**(network.n_nodes))


def orbit_length_sum_f(network):
    """
    This function takes a network object and and returns its 
    score computed by summing the lengths of genuine orbits.
    """
    binspace = range(0,num.power(2,network.n_nodes))
    genuine_orbits = []
    genuine_orbit_lengths=[]
    for state in binspace:
        (orbit_length,orbit) = network.search_equilibrium(2**network.n_nodes,state,True)
        is_in_list = False
        for old_orbit in genuine_orbits:
            for state in old_orbit:
                if all(state == orbit[-1]):
                    is_in_list = True
        if is_in_list == False:
            genuine_orbits.append(orbit)
            genuine_orbit_lengths.append(orbit_length)
    return sum(genuine_orbit_lengths)




import numpy as num
import itertools
import logging
import kreveik.classes as classes
import killer


def motif_freqs(family,degree, sorting=False, **kwargs):
    """
    Returns a list of motifs of the family.
    """
    from kreveik import network
    import copy 
    
    if  'motiflist' in kwargs:
        returned_motifs = copy.deepcopy(kwargs['motiflist'])
    else:
        returned_motifs = network.all_conn_motifs(degree)[:]

    logging.info("Computing motif frequencies of the family")
    for networkf in family:
        returned_motifs = network.motif_freqs(networkf, degree, motiflist=returned_motifs)           
        
    if sorting:
        return  sorted (returned_motifs, key = lambda returned_motifs:returned_motifs[1] , reverse = True)
    else:
        return returned_motifs

def mean_connectivity(family):
    """
    The mean connectivity of a Family of networks is returned.
    """
    [network.network for network in family]


'''
Created on Sep 8, 2011
'''

__author__ = "Mehmet Ali Anil"
__copyright__ = ""
__credits__ = ["Mehmet Ali Anil"]
__license__ = ""
__version__ = "0.0.5"
__maintainer__ = "Mehmet Ali Anil"
__email__ = "anilm@itu.edu.tr"
__status__ = "Production"
        
import numpy as num
import boolfuncs_c
import logging

def xor_masking(network,state):
    """
   
    """
    
    state = num.array(state,dtype=bool)
    newstate = num.array([None]*network.n_nodes)
    
    for i in xrange(network.n_nodes):
        #    Detect all nodes that have an incoming connection
        
        nonzero_of_adj = network.adjacency[i,].nonzero()[0]
        
        #    Reduce the boolean function and the state to a boolean function
        #     and state concerning only the incoming connections.
        
        short_mask = network.mask[i,].take(nonzero_of_adj)
        short_state = state.take(nonzero_of_adj)
    
        #    Two vectors are XOR' d element wise, and is summed in modulo 2
        #    This corresponds to i th node operating its boolean function over the same
        #    state vector.
        
        newstate[i] = (num.logical_xor(short_mask,short_state).sum()<len(short_state)/2.0)
        
    try:
        return newstate
    except:
        logging.error("XOR masking failed in network")
        logging.error("Printing id:")
        logging.error(id(network))
        return False



    
def and_masking(network,state):
    """

    """
    state = num.array(state)
    newstate=num.zeros(network.n_nodes)
    for i in range(0,network.n_nodes):
        nonzero_of_adj = network.adjacency[i,].nonzero()[0]
        short_mask = network.mask[i,].take(nonzero_of_adj)
        short_state =state.take(nonzero_of_adj)
        sum_of_bool = num.logical_and(short_mask,short_state).sum()
        newstate[i]=(len(short_state)/2.0) < sum_of_bool
    try:
        return newstate
    except:
        logging.error("AND masking failed in network")
        logging.error("Printing id:")
        logging.error(id(network))
        return False
    
def or_masking(network,state):
    """

    """
    
    newstate=num.zeros(network.n_nodes)
    for i in range(0,network.n_nodes):
        nonzero_of_adj = network.adjacency[i,].nonzero()[0]
        short_mask = network.mask[i,].take(nonzero_of_adj)
        short_state = network.state[-1].take(nonzero_of_adj)
        sum_of_bool = num.logical_or(short_mask,short_state).sum()
        newstate[i]=(len(short_state)/2.0) < sum_of_bool
    try:
        return newstate
    except:
        logging.error("OR masking failed in network")
        logging.error("Printing id:")
        logging.error(id(network))
        return False

def xor_masking_C(network,state):

    newstate = boolfuncs_c.xor_masking_c(network.adjacency,network.mask,state)
    return newstate

def or_masking_C(network,state):

    newstate = boolfuncs_c.or_masking_c(network.adjacency,network.mask,state)
    return newstate

def and_masking_C(network,state):

    newstate = boolfuncs_c.and_masking_c(network.adjacency,network.mask,state)
    return newstate



import numpy as num
import logging
from kreveik import *

def point_mutate_adj(network):
    '''
    Will result in a series of implicit point mutation of the adjacency matrix
    '''

    logging.info("Network "+str(network)+" is mutated")
    num.random.seed()
    random_i = num.random.randint(0, network.n_nodes)
    random_j = num.random.randint(0, network.n_nodes)
    network.adjacency[random_i][random_j] = not(network.adjacency[random_i][random_j])
        
def point_mutate_mask(network):
    '''
    Will result in a series of implicit point mutation of the masks
    '''                        
    
    logging.info("Network "+str(network)+" is mutated")
    num.random.seed()
    random_i = num.random.randint(0, network.n_nodes)
    random_j = num.random.randint(0, network.n_nodes)
    network.mask[random_i][random_j] = not(network.mask[random_i][random_j])
    
def degree_and_connectivity_preserving_mutation(network, maximum = 0, def_mutation = False):
    """
    A mutation that preserves the degree of the network in concern.
    """
    adj = network.adjacency
    logging.debug( "Adjacency Matrix:")
    logging.debug( adj)
    if maximum == 0:
        maximum = 10*len(adj)**2
    columns = adj.sum(axis=0)
    rows = adj.sum(axis=1)
    colsnotzero = num.where(columns != 0)[0]
    rowsnotzero = num.where(rows != 0)[0]
    #if all columns are zero or all rows are zero we can't proceed
    
    if len(colsnotzero) > 0 and len(rowsnotzero) > 0:
        for cntr in xrange(maximum):
            logging.debug( "    Initiating...")
            randomrow = rowsnotzero[num.random.randint(len(rowsnotzero))]
            randomcol = colsnotzero[num.random.randint(len(colsnotzero))]
            logging.debug("Selected item:")
            logging.debug((randomrow,randomcol))
            adj_debug = num.array(adj[:],dtype=str)
            logging.debug(adj_debug)
            adj_debug[randomrow][randomcol] = "1"
            boolean = adj[randomrow][randomcol]
            row = adj[randomrow,:]
            col = adj[:,randomcol]
            colitemsnot = num.where(col != boolean)[0]
            rowitemsnot = num.where(row != boolean)[0]
            randomrowitem = rowitemsnot[num.random.randint(len(rowitemsnot))]
            randomcolitem = colitemsnot[num.random.randint(len(colitemsnot))]
            adj_debug[randomcolitem,randomcol]="2"
            adj_debug[randomrow,randomrowitem]="3"
            adj_debug[randomcolitem,randomrowitem]="4"
            logging.debug(adj_debug)
            if adj[randomcolitem][randomrowitem] == boolean:  
                logging.debug("Success!"  )            
                adj[randomcolitem,randomrowitem] = not(adj[randomcolitem,randomrowitem])
                adj[randomcolitem,randomcol] = not(adj[randomcolitem,randomcol])
                adj[randomrow,randomrowitem] = not(adj[randomrow,randomrowitem])
                adj[randomrow,randomcol] = not(adj[randomrow,randomcol])
                logging.debug("New Network:")
                logging.debug( num.array(adj[:],dtype=str))
                if network.is_connected():
                    return network
            
        logging.info("This network "+str(network)+"is unlikely to have a degree preserving\
                      mutation. The network is returned as is.")
    else:
        logging.info("Network "+str(network)+" evolved such that there's no connection left.\
                      The network is returned as is.")

def degree_preserving_mutation(network, maximum = 0, def_mutation = False):
    """
    A mutation that preserves the degree of the network in concern.
    """
    adj = network.adjacency
    logging.debug( "Adjacency Matrix:")
    logging.debug( adj)
    if maximum == 0:
        maximum = 10*len(adj)**2
    columns = adj.sum(axis=0)
    rows = adj.sum(axis=1)
    colsnotzero = num.where(columns != 0)[0]
    rowsnotzero = num.where(rows != 0)[0]
    #if all columns are zero or all rows are zero we can't proceed
    
    if len(colsnotzero) > 0 and len(rowsnotzero) > 0:
        for cntr in xrange(maximum):
            logging.debug( "    Initiating...")
            randomrow = rowsnotzero[num.random.randint(len(rowsnotzero))]
            randomcol = colsnotzero[num.random.randint(len(colsnotzero))]
            logging.debug("Selected item:")
            logging.debug((randomrow,randomcol))
            adj_debug = num.array(adj[:],dtype=str)
            logging.debug(adj_debug)
            adj_debug[randomrow][randomcol] = "1"
            boolean = adj[randomrow][randomcol]
            row = adj[randomrow,:]
            col = adj[:,randomcol]
            colitemsnot = num.where(col != boolean)[0]
            rowitemsnot = num.where(row != boolean)[0]
            randomrowitem = rowitemsnot[num.random.randint(len(rowitemsnot))]
            randomcolitem = colitemsnot[num.random.randint(len(colitemsnot))]
            adj_debug[randomcolitem,randomcol]="2"
            adj_debug[randomrow,randomrowitem]="3"
            adj_debug[randomcolitem,randomrowitem]="4"
            logging.debug(adj_debug)
            if adj[randomcolitem][randomrowitem] == boolean:  
                logging.debug("Success!"  )            
                adj[randomcolitem,randomrowitem] = not(adj[randomcolitem,randomrowitem])
                adj[randomcolitem,randomcol] = not(adj[randomcolitem,randomcol])
                adj[randomrow,randomrowitem] = not(adj[randomrow,randomrowitem])
                adj[randomrow,randomcol] = not(adj[randomrow,randomcol])
                logging.debug("New Network:")
                logging.debug( num.array(adj[:],dtype=str))
                return network
            
        logging.info("This network "+str(network)+"is unlikely to have a degree preserving\
                      mutation. The network is returned as is.")
    else:
        logging.info("Network "+str(network)+" evolved such that there's no connection left.\
                      The network is returned as is.")
               


global JOBSERVER

def init():
    import pp
    import sys
    
    ppservers = ()
    #ppservers = ("10.0.0.1",)
    
    if len(sys.argv) > 1:
        ncpus = int(sys.argv[1])
        # Creates jobserver with ncpus workers
        
        JOBSERVER = pp.Server(ncpus, ppservers=ppservers)
        
    else:
        # Creates jobserver with automatically detected number of workers
        
        JOBSERVER = pp.Server(ppservers=ppservers)
    
    print "Starting pp with", JOBSERVER.get_ncpus(), "workers"


#from distribute_setup import use_setuptools
#use_setuptools()

from setuptools import setup, find_packages, Extension
import numpy

numpy_include = str(numpy.get_include())
module = Extension('kreveik.network.boolfuncs.boolfuncs_c', 
                   sources = ['lib/boolfuncs.c'], include_dirs=[numpy_include])

setup(
    name = "Kreveik",
    version = "0.5.6a",
    packages = find_packages(),
    install_requires = ['docutils>=0.3','numpy>=1.5','matplotlib>=1.0'],

    package_data = {
        # If any package contains *.md files, include them:
        '': ['*.md','README'],
    },

    # metadata for upload to PyPI
    author = "Mehmet Ali Anil",
    author_email = "mehmet.ali.anil@ieee.org",
    description = "Kreveik is a Python module for Boolean networks. With Kreveik, one can create, investigate dynamics of, form families of random Boolean networks.",
    license = "LICENSE.txt",
    keywords = "random boolean networks genetic algorithm",
    url = "http://github.com/mehmetalianil/Kreveik/",    
    # project home page, if any

    classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
	  'Intended Audience :: Science/Research',
          'License :: Other/Proprietary License',
	  'Natural Language :: English',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 2.7',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Physics'
          ],
      ext_modules = [module]

)



import numpy as num
import itertools
import kreveik.classes as classes
import logging
from kreveik import * 
import generators
import mutators
import scorers
import selectors
import boolfuncs

def all_conn_motifs(nodes):
    """Returns a list of all connected motifs for a degree.
    """
    logging.info("Returning all connected motifs with "+str(nodes)+" nodes")
    motiflist = []
    degree = nodes**2
    for number in xrange(2**degree):
        linear = num.array([int(item) for item in [False]*
                   (degree-len(list(bin(number)[2:])))+list(bin(number)[2:])],dtype=bool)
        motifadj = num.reshape(linear,(nodes,nodes))
        motif = classes.Motif(motifadj)
        if motif.is_connected() and not(any([motif==motiffromlist[0] for
                                              motiffromlist in motiflist])):
                motiflist.append([motif,0])
    return motiflist

    
def motif_freqs (network,degree,**kwargs):
    """
    Returns a list of motifs for a given network
    """

    logging.info("Extracting "+str(degree)+" motifs of network "+str(network))
    all_combinations = itertools.combinations(range(len(network.adjacency)),degree)
    
    if 'motiflist' in kwargs:
        allmotifs = kwargs['motiflist'][:]
        motif_list = allmotifs[:]
    else:
        logging.info("Creating all possible motifs of node count "+str(degree)+".")
        motif_list = all_conn_motifs(degree)[:]
        
    logging.info("Extracting motifs from all possible "+str(degree)+" node combinations of the network.")
    
    for combination in all_combinations:
        logging.debug("Motif Permutation:"+str(list(combination)))
        
        this_motif_adj = num.zeros((degree,degree), dtype = bool)
        for (first_ctr,first_node) in enumerate(list(combination)):
            for (second_ctr,second_node) in enumerate(list(combination)):
                this_motif_adj[first_ctr][second_ctr] = network.adjacency[first_node][second_node]
        
        this_motif = classes.Motif(this_motif_adj)
        logging.debug("Motif Adjacency:")
        logging.debug(list(combination))
        logging.debug(str(this_motif_adj))
        if this_motif.is_connected():
            truth = [this_motif == motif_vec[0] for motif_vec in motif_list]
            if (any(truth) == True):
                index = truth.index(True)
                motif_list[index][1] = motif_list[index][1]+1
            elif (all(truth) == False):
                logging.info("")
                motif_list.append([this_motif,1])
            else:
                logging.error("There has been a problem while extracting Motifs")
                break
        
    logging.info("Extraction done!")
    return motif_list

def local_clustering_in(network):
    """
    Returns the local clustering coefficient for input connections for every single node 
    """
    adj = network.adjacency*1-num.diagflat(num.diag(network.adjacency*1))
    returned = num.zeros(len(adj))
    for (counter,row) in enumerate(adj):
        k = row.sum()
        where = num.where(k)[0]
        connections=0.0
        for first in where:
            for second in where:
                if adj(first,second)==1 or adj(second,first)==1:
                    connections += 1.0
        returned[counter] = connections /((k)*(k+1))
    return returned

def local_clustering_out(network):
    """
    Returns the local clustering coefficient for output connections for every single node 
    """
    adj = num.transpose(network.adjacency*1-num.diagflat(num.diag(network.adjacency*1)))
    returned = num.zeros(len(adj))
    for (counter,row) in enumerate(adj):
        k = row.sum()
        where = num.where(k)[0]
        connections=0.0
        for first in where:
            for second in where:
                if adj(first,second)==1 or adj(second,first)==1:
                    connections += 1.0
        returned[counter] = connections /((k)*(k+1))
    return returned

def global_clustering_out(network):
    """
    Returns the global clustering coefficient for output connections for every single node 
    """
    local_clustering_out(network).mean()

def global_clustering_in(network):
    """
    Returns the global clustering coefficient for input connections for every single node 
    """
    local_clustering_in(network).mean()
                